"""

"""
from magicsoup.containers import Molecule, Chemistry

NADPH = Molecule("NADPH", 200.0 * 1e3)
NADP = Molecule("NADP", 100.0 * 1e3)
ATP = Molecule("ATP", 100.0 * 1e3)
ADP = Molecule("ADP", 70.0 * 1e3)

ammonia = Molecule("ammonia", 10.0 * 1e3)
glutamate = Molecule("glutamate", 200.0 * 1e3)
glutamine = Molecule("glutamine", 220.0 * 1e3)
oxalalcetate = Molecule("oxalalcetate", 200.0 * 1e3)

HSCoA = Molecule("HS-CoA", 200.0 * 1e3)
acetylCoA = Molecule("acetyl-CoA", 260.0 * 1e3)

MOLECULES = [
    NADPH,
    NADP,
    ATP,
    ADP,
    ammonia,
    glutamate,
    glutamine,
    oxalalcetate,
    HSCoA,
    acetylCoA,
]

REACTIONS = [
    ([glutamate, ATP, ammonia], [ADP, glutamine]),
    ([oxalalcetate, glutamine, NADPH], [glutamate, glutamate, NADP]),
]

CHEMISTRY = Chemistry(molecules=MOLECULES, reactions=REACTIONS)



import pytest
import torch
import magicsoup as ms
from magicsoup.examples.wood_ljungdahl import MOLECULES, REACTIONS
from tests.conftest import Retry


def test_molecule_amount_integrity_during_diffusion():
    chemistry = ms.Chemistry(molecules=MOLECULES, reactions=[])
    world = ms.World(chemistry=chemistry, map_size=128)

    exp = world.molecule_map.sum(dim=[1, 2])
    for step_i in range(100):
        world.diffuse_molecules()
        res = world.molecule_map.sum(dim=[1, 2])
        assert (res.sum() - exp.sum()).abs() < 10.0, step_i
        assert torch.all(torch.abs(res - exp) < 1.0), step_i


def test_molecule_amount_integrity_during_reactions():
    # X and Y can react back and forth but X + Y <-> Z
    # so if Z is counted as 2, n should stay equal
    mx = ms.Molecule("mx", 10 * 1e3)
    my = ms.Molecule("my", 20 * 1e3)
    mz = ms.Molecule("mz", 30 * 1e3)
    molecules = [mx, my, mz]
    reactions = [([mx], [my]), ([mx, my], [mz])]

    chemistry = ms.Chemistry(molecules=molecules, reactions=reactions)
    world = ms.World(chemistry=chemistry, map_size=128)
    genomes = [ms.random_genome(s=500) for _ in range(1000)]
    world.spawn_cells(genomes=genomes)

    def count(world: ms.World) -> float:
        mxy = world.molecule_map[[0, 1]].sum().item()
        mz = world.molecule_map[2].sum().item() * 2
        cxy = world.cell_molecules[:, [0, 1]].sum().item()
        cz = world.cell_molecules[:, 2].sum().item() * 2
        return mxy + mz + cxy + cz

    n0 = count(world)
    for step_i in range(100):
        world.enzymatic_activity()
        n = count(world)
        assert n == pytest.approx(n0, abs=1.0), step_i


def test_run_world_without_reactions():
    chemistry = ms.Chemistry(molecules=MOLECULES[:2], reactions=[])
    world = ms.World(chemistry=chemistry)

    genomes = [ms.random_genome(s=500) for _ in range(1000)]
    world.spawn_cells(genomes=genomes)

    for _ in range(100):
        world.enzymatic_activity()


def test_exploding_molecules():
    # if e.g. during enzyme_activity calculation a reaction is not fair,
    # it can lead to cells generating molecules from nothing
    # this happens e.g. when X.clamp(min=0.0) is done before returning
    # if one side of the equation is slowed down, the other must be too
    chemistry = ms.Chemistry(molecules=MOLECULES, reactions=REACTIONS)
    world = ms.World(chemistry=chemistry, map_size=128)
    world.spawn_cells(genomes=[ms.random_genome(s=500) for _ in range(1000)])

    for i in range(100):
        world.degrade_molecules()
        world.diffuse_molecules()
        world.enzymatic_activity()

        molmap = world.molecule_map
        cellmols = world.cell_molecules

        assert molmap.min() >= 0.0, i
        assert 0.0 < molmap.mean() < 50.0, i
        assert molmap.max() < 500.0, i

        assert cellmols.min() >= 0.0, i
        assert 0.0 < cellmols.mean() < 50.0, i
        assert cellmols.max() < 500.0, i


def test_genome_generation_consistency():
    n_tries = 5
    retry = Retry(n_allowed_fails=1)

    mi = ms.Molecule("mi", 10 * 1e3)
    mj = ms.Molecule("mj", 10 * 1e3)
    mk = ms.Molecule("mk", 10 * 1e3)
    molecules = [mi, mj, mk]
    reactions = [([mi], [mj]), ([mi, mj], [mk])]
    chemistry = ms.Chemistry(molecules=molecules, reactions=reactions)
    world = ms.World(chemistry=chemistry)

    # early stopping should not be possible
    # there is always the possibility that another protein
    # is created on the reverse-complement

    trsnp_dom = ms.TransporterDomainFact(
        molecule=mi, is_exporter=False, km=1.0, vmax=1.0
    )
    for i in range(n_tries):
        with retry.catch_assert(i):
            prtm = [ms.ProteinFact(domain_facts=trsnp_dom)]
            idxs = world.spawn_cells(
                genomes=[world.generate_genome(proteome=prtm, size=27)]
            )

            assert len(idxs) == 1
            ci = idxs[0]
            cell = world.get_cell(by_idx=ci)
            proteome = cell.get_proteome(world=world)
            assert len(proteome) == 1, proteome
            p0 = proteome[0]
            assert len(p0.domains) == 1, p0.domains
            d0 = p0.domains[0]

            assert isinstance(d0, ms.TransporterDomain)
            assert d0.molecule is mi
            assert abs(d0.vmax - 1.0) < 0.5
            assert abs(d0.km - 1.0) < 0.5
            assert not d0.is_exporter

            assert world.kinetics.N[ci][0][0] == 1, world.kinetics.N[ci]
            assert world.kinetics.N[ci][0][3] == -1, world.kinetics.N[ci]
            assert abs(world.kinetics.Vmax[ci][0] - 1.0) < 0.5
            assert abs(world.kinetics.Kmf[ci][0] - 1.0) < 0.5
            assert abs(world.kinetics.Kmb[ci][0] - 1.0) < 0.5

    world.kill_cells(cell_idxs=list(range(world.n_cells)))
    retry.reset()

    catal_dom = ms.CatalyticDomainFact(reaction=([mj], [mi]), km=1.0, vmax=1.0)
    for i in range(n_tries):
        with retry.catch_assert(i):
            prtm = [ms.ProteinFact(domain_facts=catal_dom)]
            idxs = world.spawn_cells(
                genomes=[world.generate_genome(proteome=prtm, size=27)]
            )

            assert len(idxs) == 1
            ci = idxs[0]
            cell = world.get_cell(by_idx=ci)
            proteome = cell.get_proteome(world=world)
            assert len(proteome) == 1, proteome
            p0 = proteome[0]
            assert len(p0.domains) == 1
            d0 = p0.domains[0]

            assert isinstance(d0, ms.CatalyticDomain)
            assert d0.substrates[0] is mj
            assert d0.products[0] is mi
            assert abs(d0.vmax - 1.0) < 0.5
            assert abs(d0.km - 1.0) < 0.5

            assert world.kinetics.N[ci][0][0] == 1, world.kinetics.N[ci]
            assert world.kinetics.N[ci][0][1] == -1, world.kinetics.N[ci]
            assert abs(world.kinetics.Vmax[ci][0] - 1.0) < 0.5
            assert abs(world.kinetics.Kmf[ci][0] - 1.0) < 0.5
            assert abs(world.kinetics.Kmb[ci][0] - 1.0) < 0.5

    world.kill_cells(cell_idxs=list(range(world.n_cells)))
    retry.reset()

    # with 2 domains there is a possibility that the last codon of the
    # 1st domain is a stop codon (=> a protein with only the 1st domain)
    # additionally, there could be a start codon in the first domain
    # (=> additional protein with only the second domain)
    # the reverse complement can always harbour another protein

    reg_dom = ms.RegulatoryDomainFact(
        effector=mk, is_transmembrane=True, is_inhibiting=True, km=1.0
    )
    for i in range(n_tries):
        with retry.catch_assert(i):
            prtm = [ms.ProteinFact(domain_facts=[reg_dom, catal_dom])]
            idxs = world.spawn_cells(
                genomes=[world.generate_genome(proteome=prtm, size=48)]
            )

            assert len(idxs) == 1
            ci = idxs[0]
            cell = world.get_cell(by_idx=ci)
            proteome = cell.get_proteome(world=world)
            assert len(proteome) >= 1, proteome
            p0 = proteome[0]
            assert len(p0.domains) >= 2, p0.domains

            i = 0 if isinstance(p0.domains[0], ms.RegulatoryDomain) else 1
            assert isinstance(p0.domains[i], ms.RegulatoryDomain)
            assert p0.domains[i].effector is mk
            assert abs(p0.domains[i].km - 1.0) < 0.5
            assert p0.domains[i].is_inhibiting
            assert p0.domains[i].is_transmembrane
            assert world.kinetics.Ai[ci][0][5] == 1, world.kinetics.Ai[ci]
            assert abs(world.kinetics.Kmr[ci][0] - 1.0) < 0.5



"""
# CO2 fixation pathways

- Calvin cycle
- 3-hydroxypropionate cycle
- Wood-Ljungdahl cycle
- reductive TCA cycle
- dicarboxylate/4-hydroxybutyrate cycle
- 3-hydroxypropionate/4-hydroxybutyrate cycle

from Gong, Fuyu & Cai, Zhen & Li, Yin. (2016). Synthetic biology for CO2 fixation.
and https://en.wikipedia.org/wiki/Biological_carbon_fixation#Overview%20of%20pathways

Some reactions are combined for simplicity.
Common intermediates between different pathways are kept
to allow cell to use combinations of pathways.

## Calvin cycle

Stage 1:
RuBP + CO2 -> 2 3PGA (-35 kJ/mol)
3PGA + ATP -> 1,3BPG + ADP
1,3BPG + NADPH -> G3P + NADP (-21.9 kJ/mol)

Stage 2:
5 G3P -> 3 Ru5P
Ru5P + ATP -> RuBP + ADP

essential: RuBP

## 3-Hydroxypropionate Bicycle

CO2 fixation:
acetyl-CoA + CO2 + ATP -> malonyl-CoA + ADP
malonyl-CoA + 3 NADPH + ATP -> propionyl-CoA + 3 NADP + AMP
propionyl-CoA + CO2 + ATP -> methylmalonyl-CoA + ADP
methylmalonyl-CoA -> succinyl-CoA
succinyl-CoA -> succinate + HS-CoA
succinate + NADP -> fumarate + NADPH
fumarate -> malate
malate + HS-CoA -> malyl-CoA
malyl-CoA -> acetyl-CoA + glyoxylate

Glyoxylate Assimilation (-14 kJ/mol):
propionyl-CoA + glyoxylate -> methylmalyl-CoA
methylmalyl-CoA -> citramalyl-CoA
citramalyl-CoA -> acetyl-CoA + pyruvate

essential: HS-CoA

## Wood-Ljungdahl cycle

Methyl (Eastern) Branch:
CO2 + NADPH -> formate + NADP (- 21.5 kJ/mol)
formate + FH4 + ATP -> formyl-FH4 + ADP (-8.4 kJ/mol)
formyl-FH4 + NADPH -> methylen-FH4 + NADP
methylen-FH4 + NADPH -> methyl-FH4 + NADP (-57.3 kJ/mol)

Carbonyl (Western) Branch:
CO2 + 2 Fd-red -> CO + 2 Fd-ox
methyl-FH4 + CO + HS-CoA -> acetyl-CoA + FH4

essential: FH4, HS-CoA

## Reductive TCA

oxalacetate + NADPH -> malate + NADP (-64.2 kJ/mol)
malate -> fumarate (+2.3 kJ/mol)
fumarate + FADH -> succinate + FAD (-102.2 kJ/mol)
succinate + HS-CoA + ATP -> succinyl-CoA + ADP
succinyl-CoA + FdRed + CO2 -> a-ketoglutarate + HS-CoA + FdOx (-14.8 kJ/mol incl prev)
a-ketoglutarate + CO2 + NADPH -> isocitrate + NADP (+3.8 kJ/mol)
isocitrate -> citrate (-8.5 kJ/mol)
citrate + HS-CoA + ATP -> oxalacetate + acetyl-CoA + ADP

essential: HS-CoA, oxalacetate (could also be created from pyruvate)

## Dicarboxylate/4-Hydroxybutyrate Cycle

acetyl-CoA + Fd-red -> pyruvate + Fd-ox + HS-CoA
pyruvate + ATP -> PEP + AMP
PEP + CO2 -> oxalacetate
oxalacetate + NADPH -> malate + NADP
malate -> fumarate
fumarate + 2 MV-red -> succinate + 2 MV-ox
succinate + HS-CoA + ATP -> succinyl-CoA + ADP
succinyl-CoA + 2 MV-red -> succinic semialdehyde + HS-CoA + 2 MV-ox
succinic semialdehyde + NADPH -> 4-hydroxy-butyrate + NADP
4-hydroxy-butyrate + HS-CoA + ATP -> hydroxybutyryl-CoA + AMP
hydroxybutyryl-CoA + NADP -> acetoacetyl-CoA + NADPH
acetoacetyl-CoA + HS-CoA -> 2 acetyl-CoA

essential: HS-CoA

## 3-Hydroxypropionate/4-Hydroxybutyrate Cycle

acetyl-CoA + CO2 + ATP -> malonyl-CoA + ADP
malonyl-CoA + 3 NADPH + ATP -> propionyl-CoA + 3 NADP + ADP
propionyl-CoA + CO2 + ATP -> methylmalonyl-CoA + ADP
methylmalonyl-CoA -> succinyl-CoA
succinyl-CoA + NADPH -> succinic semialdehyde + HS-CoA + NADP
succinic semialdehyde + NADPH -> 4-hydroxy-butyrate + NADP
4-hydroxy-butyrate + HS-CoA + ATP -> hydroxybutyryl-CoA + ADP
hydroxybutyryl-CoA + NADP -> acetoacetyl-CoA + NADPH
acetoacetyl-CoA + HS-CoA -> 2 acetyl-CoA

essential: HS-CoA

## Common Reactions

Some reactions are powered by ATP or NADPH.
Generally, cells should create energy coupling themselves.
So, the reaction is defined without ATP/NADPH.

ATP -> ADP (-30.5 kJ/mol)
NADPH -> NADP (-73.2 kcal/mol)

But in some cases the phosphorylation or electron transfer is important.
In these cases, the reaction will be defined together with ATP/NADPH.
Here, ATP -> ADP is the representative energy carrier and phosphate donor.
E.g. ADP -> AMP is not defined.
NADPH is the representative energy carrier and electron donor.
E.g. no FADH2 or Fd-red is defined.

Molecule X will be used to caputure biologically available carbon.
Molecule E will be used to replenish energy carriers.

## Finding Energies

1. decided to have NADPH->NADP (-70), ATP->ADP (-30) to match real data
2. decided on some base energy for carriers (e.g. HS-CoA=200)
3. for each metabolite counted number of Ps, Cs, and C-C/C-H bonds
4. created initial energies for metabolites by multiplying these counts:
   20 * [P] + 10 * [C] + 60 * [C-C/C-H]
5. started adjusting these initial energies to yield the real reaction
   energies I found in publications (e.g. in WL and reductive TCA pathway)
6. result is a tradeoff by iterative trial-and-error to come close to
   real reaction energies but also be consistent across all pathways
7. derived energies and reactions for X and E to be as fair as possible
   for any carrier molecule

I would like to add more transformations that are not described in these
pathway but I don't know enough about this topic.
E.g. shoudln't there be a hydroxybutyrate <-> malate (or similar)?
"""
from magicsoup.containers import (
    Molecule,
    Chemistry,
)

_TReact = tuple[list[Molecule], list[Molecule]]

_co2 = Molecule("CO2", 10.0 * 1e3, diffusivity=1.0, permeability=1.0)
_NADPH = Molecule("NADPH", 200.0 * 1e3)
_NADP = Molecule("NADP", 130.0 * 1e3)
_ATP = Molecule("ATP", 100.0 * 1e3)
_ADP = Molecule("ADP", 65.0 * 1e3)
_G3P = Molecule("G3P", 420.0 * 1e3)  # P 3C 12e
_acetylCoA = Molecule("acetyl-CoA", 475.0 * 1e3)  # 2C 8e
_HSCoA = Molecule("HS-CoA", 190.0 * 1e3)
_pyruvate = Molecule("pyruvate", 330.0 * 1e3)  # 3C 10e
_X = Molecule("X", 50.0 * 1e3)
_E = Molecule("E", 150.0 * 1e3)

_common_reacts: list[_TReact] = [
    # energy carriers with defined reaction energies
    ([_NADPH], [_NADP]),  # -70
    ([_ATP], [_ADP]),  # -35
    # restoring them should be practically irreversible
    ([_ADP, _ADP, _E], [_ATP, _ATP]),  # -80
    ([_NADP, _E], [_NADPH]),  # -80
    # extracting carbon should be as fair as possible for all
    ([_G3P], [_X, _X, _X, _X, _X, _X, _X, _X]),  # -20
    ([_pyruvate], [_X, _X, _X, _X, _X, _X]),  # -30
    ([_acetylCoA], [_HSCoA, _X, _X, _X, _X, _X]),  # -35
]

_common_mols = [
    _co2,
    _NADPH,
    _NADP,
    _ATP,
    _ADP,
    _acetylCoA,
    _HSCoA,
    _pyruvate,
    _G3P,
    _X,
    _E,
]

# Calvin
_RuBP = Molecule("RuBP", 725.0 * 1e3)  # 2P 5C 20e
_3PGA = Molecule("3PGA", 350.0 * 1e3)  # P 3C 10e
_13BPG = Molecule("1,3BPG", 370.0 * 1e3)  # 2P 3C 10e
_G3P = Molecule("G3P", 420.0 * 1e3)  # P 3C 12e
_Ru5P = Molecule("Ru5P", 695.0 * 1e3)  # P 5C 20e

_calvin_reacts = [
    ([_RuBP, _co2], [_3PGA, _3PGA]),  # -35
    ([_3PGA, _ATP], [_13BPG, _ADP]),  # -15
    ([_13BPG, _NADPH], [_G3P, _NADP]),  # -20
    ([_G3P, _G3P, _G3P, _G3P, _G3P], [_Ru5P, _Ru5P, _Ru5P]),  # -15
    ([_Ru5P, _ATP], [_RuBP, _ADP]),  # -5
]

_calvin_mols = [_RuBP, _3PGA, _13BPG, _G3P, _Ru5P]


# 3-Hydroxypropionate bicycle
_acetylCoA = Molecule("acetyl-CoA", 475.0 * 1e3)  # 2C 8e !
_HSCoA = Molecule("HS-CoA", 190.0 * 1e3)
_malonylCoA = Molecule("malonyl-CoA", 495.0 * 1e3)  # 3C 8e !
_propionylCoA = Molecule("propionyl-CoA", 675.0 * 1e3)  # 3C 14e !
_methylmalonylCoA = Molecule("methylmalonyl-CoA", 685.0 * 1e3)  # 4C 14e !
_succinylCoA = Molecule("succinyl-CoA", 685.0 * 1e3)  # 4C 14e !
_succinate = Molecule("succinate", 485.0 * 1e3)  # 4C 14e !
_fumarate = Molecule("fumarate", 415.0 * 1e3)  # 4C 12e !
_malate = Molecule("malate", 415.0 * 1e3)  # 4C 12e !
_malylCoA = Molecule("malyl-CoA", 615.0 * 1e3)  # 4C 12e !
_glyoxylate = Molecule("glyoxylate", 140.0 * 1e3)  # 2C 4e !
_methylmalylCoA = Molecule("methylmalyl-CoA", 810.0 * 1e3)  # 5C 18e !
_citramalylCoA = Molecule("citramalyl-CoA", 810.0 * 1e3)  # 5C 18e !
_pyruvate = Molecule("pyruvate", 330.0 * 1e3)  # 3C 10e !

_hprop_reacts = [
    ([_acetylCoA, _co2], [_malonylCoA]),  # +10
    (
        [_malonylCoA, _NADPH, _NADPH, _NADPH],
        [_propionylCoA, _NADP, _NADP, _NADP],
    ),  # -30
    ([_propionylCoA, _co2], [_methylmalonylCoA]),  # 0
    ([_methylmalonylCoA], [_succinylCoA]),  # 0
    ([_succinylCoA], [_succinate, _HSCoA]),  # -10
    ([_succinate, _NADP], [_fumarate, _NADPH]),  # 0
    ([_fumarate], [_malate]),  # 0
    ([_malate, _HSCoA], [_malylCoA]),  # +10
    ([_malylCoA], [_acetylCoA, _glyoxylate]),  # 0
    ([_propionylCoA, _glyoxylate], [_methylmalylCoA]),  # -5
    ([_methylmalylCoA], [_citramalylCoA]),  # 0
    ([_citramalylCoA], [_acetylCoA, _pyruvate]),  # -5
]

_hprop_mols = [
    _malonylCoA,
    _propionylCoA,
    _methylmalonylCoA,
    _succinylCoA,
    _succinate,
    _fumarate,
    _malate,
    _malylCoA,
    _glyoxylate,
    _methylmalylCoA,
    _citramalylCoA,
    _pyruvate,
]

# WL
_methylFH4 = Molecule("methyl-FH4", 410.0 * 1e3)  # 1C 6e !
_methylenFH4 = Molecule("methylen-FH4", 355.0 * 1e3)  # 1C 4e !
_formylFH4 = Molecule("formyl-FH4", 295.0 * 1e3)  # 1C 2e !
_FH4 = Molecule("FH4", 200.0 * 1e3)  # !
_formate = Molecule("formate", 70.0 * 1e3)  # 1C 2e !
_co = Molecule("CO", 75 * 1e3, diffusivity=1.0, permeability=1.0)  # 1C 2e !
_acetylCoA = Molecule("acetyl-CoA", 475.0 * 1e3)  # 2C 8e !
_HSCoA = Molecule("HS-CoA", 190.0 * 1e3)

_wl_reacts = [
    ([_co2, _NADPH], [_formate, _NADP]),  # -10
    ([_formate, _FH4], [_formylFH4]),  # -10
    ([_formylFH4, _NADPH], [_methylenFH4, _NADP]),  # -10
    ([_methylenFH4, _NADPH], [_methylFH4, _NADP]),  # -15
    ([_co2, _NADPH], [_co, _NADP]),  # -5
    ([_methylFH4, _co, _HSCoA], [_acetylCoA, _FH4]),  # 0
]

_wl_mols = [
    _methylFH4,
    _methylenFH4,
    _formylFH4,
    _FH4,
    _formate,
    _co,
]

# Reductive TCA
_acetylCoA = Molecule("acetyl-CoA", 475.0 * 1e3)  # 2C 8e !
_HSCoA = Molecule("HS-CoA", 190.0 * 1e3)
_oxalacetate = Molecule("oxalacetate", 350.0 * 1e3)  # 4C 10e !
_malate = Molecule("malate", 415.0 * 1e3)  # 4C 12e !
_fumarate = Molecule("fumarate", 415.0 * 1e3)  # 4C 12e !
_succinate = Molecule("succinate", 485.0 * 1e3)  # 4C 14e !
_succinylCoA = Molecule("succinyl-CoA", 685.0 * 1e3)  # 4C 14e !
_aKetoglutarate = Molecule("alpha-ketoglutarate", 540.0 * 1e3)  # 5C 16e !
_isocitrate = Molecule("isocitrate", 600.0 * 1e3)  # 5C 18e !
_citrate = Molecule("citrate", 600.0 * 1e3)  # 5C 18e !

_rtca_reacts = [
    ([_oxalacetate, _NADPH], [_malate, _NADP]),  # -5
    ([_malate], [_fumarate]),  # 0
    ([_fumarate, _NADPH], [_succinate, _NADP]),  # 0
    ([_succinate, _HSCoA], [_succinylCoA]),  # +10
    ([_succinylCoA, _NADPH, _co2], [_aKetoglutarate, _HSCoA, _NADP]),  # -35
    ([_aKetoglutarate, _co2, _NADPH], [_isocitrate, _NADP]),  # -20
    ([_isocitrate], [_citrate]),  # 0
    ([_citrate, _HSCoA], [_oxalacetate, _acetylCoA]),  # +35
]

_rtca_mols = [
    _oxalacetate,
    _malate,
    _fumarate,
    _succinate,
    _succinylCoA,
    _aKetoglutarate,
    _isocitrate,
    _citrate,
]

# Dicarboxylate/4-Hydroxybutyrate Cycle
_acetylCoA = Molecule("acetyl-CoA", 475.0 * 1e3)  # 2C 8e !
_pyruvate = Molecule("pyruvate", 330.0 * 1e3)  # 3C 10e !
_PEP = Molecule("PEP", 350.0 * 1e3)  # 1P 3C 10e !
_oxalacetate = Molecule("oxalacetate", 350.0 * 1e3)  # 4C 10e !
_malate = Molecule("malate", 415.0 * 1e3)  # 4C 12e !
_fumarate = Molecule("fumarate", 415.0 * 1e3)  # 4C 12e !
_succinate = Molecule("succinate", 485.0 * 1e3)  # 4C 14e !
_succinylCoA = Molecule("succinyl-CoA", 685.0 * 1e3)  # 4C 14e !
_SSA = Molecule("SSA", 535.0 * 1e3)  # 4C 16e !
_GHB = Molecule("GHB", 600.0 * 1e3)  # 4C 18e !
_hydroxybutyrylCoA = Molecule("hydroxybutyryl-CoA", 825.0 * 1e3)
_acetoacetylCoA = Molecule("acetoacetyl-CoA", 760.0 * 1e3)  # 4C 16e !

_dcarbhb_reacts = [
    ([_acetylCoA, _co2, _NADPH], [_pyruvate, _HSCoA, _NADP]),  # -35
    ([_pyruvate, _ATP], [_PEP, _ADP]),  # -15
    ([_PEP, _co2], [_oxalacetate]),  # -10
    ([_oxalacetate, _NADPH], [_malate, _NADP]),  # -5
    ([_malate], [_fumarate]),  # 0
    ([_fumarate, _NADPH], [_succinate, _NADP]),  # 0
    ([_succinate, _HSCoA], [_succinylCoA]),  # +10
    ([_succinylCoA, _NADPH], [_SSA, _HSCoA, _NADP]),  # -30
    ([_SSA, _NADPH], [_GHB, _NADP]),  # -5
    ([_GHB, _HSCoA], [_hydroxybutyrylCoA]),  # +35
    ([_hydroxybutyrylCoA, _NADP], [_acetoacetylCoA, _NADPH]),  # +5
    ([_acetoacetylCoA, _HSCoA], [_acetylCoA, _acetylCoA]),  # 0
]

_dcarbhb_mols = [
    _pyruvate,
    _PEP,
    _oxalacetate,
    _malate,
    _fumarate,
    _succinate,
    _succinylCoA,
    _SSA,
    _GHB,
    _hydroxybutyrylCoA,
    _acetoacetylCoA,
]

# 3-Hydroxypropionate/4-Hydroxybutyrate Cycle
_acetylCoA = Molecule("acetyl-CoA", 475.0 * 1e3)
_malonylCoA = Molecule("malonyl-CoA", 495.0 * 1e3)
_propionylCoA = Molecule("propionyl-CoA", 675.0 * 1e3)
_methylmalonylCoA = Molecule("methylmalonyl-CoA", 685.0 * 1e3)
_succinylCoA = Molecule("succinyl-CoA", 685.0 * 1e3)
_SSA = Molecule("SSA", 535.0 * 1e3)
_GHB = Molecule("GHB", 600.0 * 1e3)
_hydroxybutyrylCoA = Molecule("hydroxybutyryl-CoA", 825.0 * 1e3)
_acetoacetylCoA = Molecule("acetoacetyl-CoA", 760.0 * 1e3)

_hprophbut_reacts = [
    ([_acetylCoA, _co2], [_malonylCoA]),  # +10
    (
        [_malonylCoA, _NADPH, _NADPH, _NADPH],
        [_propionylCoA, _NADP, _NADP, _NADP],
    ),  # -30
    ([_propionylCoA, _co2], [_methylmalonylCoA]),  # 0
    ([_methylmalonylCoA], [_succinylCoA]),  # 0
    ([_succinylCoA, _NADPH], [_SSA, _HSCoA, _NADP]),  # -30
    ([_SSA, _NADPH], [_GHB, _NADP]),  # -5
    ([_GHB, _HSCoA], [_hydroxybutyrylCoA]),  # +35
    ([_hydroxybutyrylCoA, _NADP], [_acetoacetylCoA, _NADPH]),  # +5
    ([_acetoacetylCoA, _HSCoA], [_acetylCoA, _acetylCoA]),  # 0
]

_hprophbut_mols = [
    _malonylCoA,
    _propionylCoA,
    _methylmalonylCoA,
    _succinylCoA,
    _SSA,
    _GHB,
    _hydroxybutyrylCoA,
    _acetoacetylCoA,
]

# combined
MOLECULES = (
    _common_mols
    + _calvin_mols
    + _wl_mols
    + _hprop_mols
    + _rtca_mols
    + _dcarbhb_mols
    + _hprophbut_mols
)

REACTIONS = (
    _common_reacts
    + _calvin_reacts
    + _wl_reacts
    + _hprop_reacts
    + _rtca_reacts
    + _dcarbhb_reacts
    + _hprophbut_reacts
)

CHEMISTRY = Chemistry(molecules=MOLECULES, reactions=REACTIONS)



from typing import Any
import math
import random
import torch
from magicsoup.constants import GAS_CONSTANT, DomainSpecType, ProteinSpecType
from magicsoup.containers import (
    Molecule,
    Protein,
    CatalyticDomain,
    TransporterDomain,
    RegulatoryDomain,
    Domain,
)

# MAX,MIN should be at least x100 away from inf
# EPS should be 1/MAX
_EPS = 1e-36
_MAX = 1e36
_MIN = -1e36


class _LogNormWeightMapFact:
    """
    Creates an object that maps tokens to a float
    which is sampled from a log normal distribution.
    """

    def __init__(
        self,
        max_token: int,
        weight_range: tuple[float, float],
        device: str = "cpu",
        zero_value: float = torch.nan,
    ):
        min_w = min(weight_range)
        max_w = max(weight_range)
        l_min_w = math.log(min_w)
        l_max_w = math.log(max_w)
        mu = (l_min_w + l_max_w) / 2
        sig = l_max_w - l_min_w
        non_zero_weights: list[float] = []
        for _ in range(max_token):
            sample = math.exp(random.gauss(mu, sig))
            while not min_w <= sample <= max_w:
                sample = math.exp(random.gauss(mu, sig))
            non_zero_weights.append(sample)
        weights = torch.tensor([zero_value] + non_zero_weights)
        self.weights = weights.to(device)

    def __call__(self, t: torch.Tensor) -> torch.Tensor:
        """t: long (c, p, d)"""
        return self.weights[t]

    def inverse(self) -> dict[float, list[int]]:
        flt_map: dict[float, list[int]] = {}
        M = self.weights.to("cpu")
        for i in range(1, M.size(0)):
            v = M[i].item()
            if v not in flt_map:
                flt_map[v] = []
            flt_map[v].append(i)
        return flt_map


class _SignMapFact:
    """
    Creates an object that maps tokens to 1.0 or -1.0
    with 50% probability of each being mapped.
    """

    def __init__(self, max_token: int, device: str = "cpu", zero_value: float = 0.0):
        choices = [1.0, -1.0]
        signs = torch.tensor([zero_value] + random.choices(choices, k=max_token))
        self.signs = signs.to(device)

    def __call__(self, t: torch.Tensor) -> torch.Tensor:
        """t: long (c, p, d)"""
        return self.signs[t]

    def inverse(self) -> dict[bool, list[int]]:
        sign_map = {}
        M = self.signs.to("cpu")
        sign_map[True] = torch.argwhere(M == 1.0).flatten().tolist()
        sign_map[False] = torch.argwhere(M == -1.0).flatten().tolist()
        return sign_map


class _VectorMapFact:
    """
    Create an object that maps tokens
    to a list of vectors. Each vector will be mapped with
    the same frequency.
    """

    def __init__(
        self,
        max_token: int,
        n_signals: int,
        vectors: list[list[float]],
        device: str = "cpu",
        zero_value: float = 0.0,
    ):
        n_vectors = len(vectors)
        M = torch.full((max_token + 1, n_signals), fill_value=zero_value)

        if n_vectors == 0:
            self.M = M.to(device)
            return

        if not all(len(d) == n_signals for d in vectors):
            raise ValueError(f"Not all vectors have length of signal_size={n_signals}")

        if n_vectors > max_token:
            raise ValueError(
                f"There are max_token={max_token} and {n_vectors} vectors."
                " It is not possible to map all vectors"
            )

        for vector in vectors:
            if all(d == 0.0 for d in vector):
                raise ValueError(
                    "At least one vector includes only zeros."
                    " Each vector should contain at least one non-zero value."
                )

        idxs = random.choices(list(range(n_vectors)), k=max_token)
        for row_i, idx in enumerate(idxs):
            M[row_i + 1] = torch.tensor(vectors[idx])
        self.M = M.to(device)

    def __call__(self, t: torch.Tensor) -> torch.Tensor:
        """t: long (c, p, d)"""
        return self.M[t]


class _ReactionMapFact(_VectorMapFact):
    """
    Create an object that maps tokens to vectors.
    Each vector has signals length and represents the
    stoichiometry of a reaction.
    """

    def __init__(
        self,
        molmap: dict[Molecule, int],
        reactions: list[tuple[list[Molecule], list[Molecule]]],
        max_token: int,
        device: str = "cpu",
        zero_value: float = 0.0,
    ):
        n_signals = 2 * len(molmap)
        n_reacts = len(reactions)

        # careful, only copy [0] to avoid having references to the same list
        vectors = [[0.0] * n_signals for _ in range(n_reacts)]
        for ri, (lft, rgt) in enumerate(reactions):
            for mol in lft:
                mol_i = molmap[mol]
                vectors[ri][mol_i] -= 1
            for mol in rgt:
                mol_i = molmap[mol]
                vectors[ri][mol_i] += 1

        super().__init__(
            vectors=vectors,
            n_signals=n_signals,
            max_token=max_token,
            device=device,
            zero_value=zero_value,
        )

    def inverse(
        self,
        molmap: dict[Molecule, int],
        reactions: list[tuple[list[Molecule], list[Molecule]]],
        n_signals: int,
    ) -> dict[tuple[tuple[Molecule, ...], tuple[Molecule, ...]], list[int]]:
        react_map = {}
        M = self.M.to("cpu")
        for subs, prods in reactions:
            t = torch.zeros(n_signals)
            for sub in subs:
                t[molmap[sub]] -= 1
            for prod in prods:
                t[molmap[prod]] += 1
            idxs = torch.argwhere((M == t).all(dim=1)).flatten().tolist()
            react_map[(tuple(subs), tuple(prods))] = idxs
        return react_map


class _TransporterMapFact(_VectorMapFact):
    """
    Create an object that maps tokens to vectors.
    Each vector has signals length and represents the
    stoichiometry of a molecule transport into or out of the cell.
    """

    def __init__(
        self,
        n_molecules: int,
        max_token: int,
        device: str = "cpu",
        zero_value: float = 0.0,
    ):
        n_signals = 2 * n_molecules

        # careful, only copy [0] to avoid having references to the same list
        vectors = [[0.0] * n_signals for _ in range(n_molecules)]
        for mi in range(n_molecules):
            vectors[mi][mi] = -1
            vectors[mi][mi + n_molecules] = 1

        super().__init__(
            vectors=vectors,
            n_signals=n_signals,
            max_token=max_token,
            device=device,
            zero_value=zero_value,
        )

    def inverse(self, molecules: list[Molecule]) -> dict[Molecule, list[int]]:
        trnsp_map = {}
        M = self.M.to("cpu")
        for mi, mol in enumerate(molecules):
            idxs = torch.argwhere(M[:, mi] != 0).flatten().tolist()
            trnsp_map[mol] = idxs
        return trnsp_map


class _RegulatoryMapFact(_VectorMapFact):
    """
    Create an object that maps tokens to vectors.
    Each vector has signals length and represents the
    either activating (+1) or inhibiting (-1) effect
    of an effector molecule.
    """

    def __init__(
        self,
        n_molecules: int,
        max_token: int,
        device: str = "cpu",
        zero_value: float = 0.0,
    ):
        n_signals = 2 * n_molecules

        # careful, only copy [0] to avoid having references to the same list
        vectors = [[0.0] * n_signals for _ in range(n_signals)]
        for mi in range(n_signals):
            vectors[mi][mi] = 1

        super().__init__(
            vectors=vectors,
            n_signals=n_signals,
            max_token=max_token,
            device=device,
            zero_value=zero_value,
        )

    def inverse(
        self, molecules: list[Molecule]
    ) -> dict[tuple[Molecule, bool], list[int]]:
        n = len(molecules)
        reg_map = {}
        M = self.M.to("cpu")
        for mi, mol in enumerate(molecules):
            idxs_int = torch.argwhere(M[:, mi] != 0).flatten().tolist()
            idxs_ext = torch.argwhere(M[:, mi + n] != 0).flatten().tolist()
            reg_map[(mol, False)] = idxs_int
            reg_map[(mol, True)] = idxs_ext
        return reg_map


class Kinetics:
    """
    Class holding logic for simulating protein work.
    Usually this class is instantiated automatically when initializing [World][magicsoup.world.World].
    You can access it on `world.kinetics`.

    Arguments:
        molecules: List of molecule species.
            They have to be in the same order as they are on `chemistry.molecules`.
        reactions: List of all possible reactions in this simulation as a list of tuples: `(substrates, products)`.
            All reactions are reversible and happen in both directions (left to right or vice versa).
        abs_temp: Absolute temperature in Kelvin will influence the free Gibbs energy calculation of reactions.
            Higher temperature will give the reaction quotient term higher importance.
        km_range: The range from which to sample Michaelis Menten constants for domains (in mM).
            They are sampled from a lognormal distribution, so all values must be > 0.
        vmax_range: The range from which to sample maximum velocities for domains (in mM/s).
            They are sampled from a lognormal distribution, so all values must be > 0.
        device: Device to use for tensors
            (see [pytorch CUDA semantics](https://pytorch.org/docs/stable/notes/cuda.html)).
            This has to be the same device that is used by `world`.
        scalar_enc_size: Number of tokens that can be used to encode the scalars Vmax, Km, and sign.
            This should be the output of `max(genetics.one_codon_map.values())`.
        vector_enc_size: Number of tokens that can be used to encode the vectors for reactions and molecules.
            This should be the output of `max(genetics.two_codon_map.values())`.

    There are `c` cells, `p` proteins, `s` signals.
    Signals are basically molecule species, but we have to differentiate between intra- and extracellular molecule species.
    So, there are twice as many signals as molecule species.
    The order of molecule species is always the same as in `chemistry.molecules`.
    First, all intracellular molecule species are listed, then all extracellular.
    The order of cells is always the same as in `world.cell_genomes` and the order of proteins
    for every cell is always the same as the order of proteins in a cell object `cell.get_proteome(world=world)`.

    Attributes on this class describe cell parameters:

    - `Kmf`, `Kmb`, `Kmr` Affinities to all signals processed by each protein in every cell (c, p).
      There are affinities for (f)orward and (b)ackward reactions, and for (r)egulatory domains.
    - `Vmax` Maximum velocities of every protein in every cell (c, p).
    - `E` Standard reaction Gibbs free energy of every protein in every cell (c, p).
    - `N` Stoichiometric number for every signal that is processed by each protein in every cell (c, p, s).
      Additionally, there are `Nf` and `Nb` which describe only forward and backward stoichiometric numbers.
      This is needed in addition to `N` to properly describe reactions that involve molecules
      which are not changed, _i.e._ where `n=0`.
    - `Ai`, `Aa` Regulatory effect for each signal in every protein in every cell (c, p, s).
      `Ai` describes inhibitors, `Aa` describes activators with numbers > 0.0 describing
      the stoichiometric number of the effector.

    The main method is [integrate_signals()][magicsoup.kinetics.Kinetics].
    When calling [enzymatic_activity()][magicsoup.world.World.enzymatic_activity], a matrix `X` of signals (c, s) is prepared
    and then [integrate_signals(X)][magicsoup.kinetics.Kinetics] is called.
    Updated signals are returned and [World][magicsoup.world.World] writes them back to `world.cell_molecules` and `world.molecule_map`.

    Another method, which ended up here, is [set_cell_params()][magicsoup.kinetics.Kinetics.set_cell_params]
    which reads proteomes and updates cell parameters accordingly.
    This is called whenever the proteomes of some cells changed.
    Currently, this is also the main performance bottleneck.

    When this class is initialized it generates the mappings from nucleotide sequences to domains by random sampling.
    These mappings are then used throughout the simulation.
    If you initialize this class again, these mappings will be different.
    Initializing [World][magicsoup.world.World] will also create one `Kinetics` instance. It is on `world.kinetics`.
    If you want to access nucleotide to domain mappings of your simulation, you should use `world.kinetics`.

    Note: All default values are based on the assumption that energies are in J, a time step represents 1s,
    and molecule numbers are in mM
    If you change the defaults, you might reconsider how these numbers should be interpreted.
    If you lower the minimum `Km` or raise the maximum `Vmax`, please also note the comments about
    `n_computation` and `alpha` below.

    The kinetics used here can never create negative molecule concentrations and make the reaction quotient move
    towards to equilibrium constant at all times.
    However, this simulation computes these things one step at a time
    (with the premise that one step should be something similar to 1s).
    This and the numerical limits of data types used here can create edge cases that need to be dealth with:
    reaction quotients overshooting the equilibrium state and negative concentrations.
    Both are caused by high $V_{max}$ with low $K_m$ values.

    If an enzyme is far away from its equilibrium state $K_e$ and substrate concentrations are far above $K_M$ it
    will progress its reaction at full speed $V_{max}$. This rate can be so high that, within one step, $Q$ surpasses
    $K_E$. In the next step it will move at full speed into the opposite direction, overshooting $K_E$ again, and so on.
    Reactions with high stoichiometric numbers are more prone to this as their rate functions are sharper.
    To combat this [integrate_signals()][magicsoup.kinetics.Kinetics] works by iteratively approaching
    $V_{max}$ on multiple levels of the computation.
    The approach was tuned to compute fast and give satifisfying reults with certain conditions in mind.
    $V_{max} \le 100$, $K_M \ge 0.01$, and concentrations $X$ generally not much higher than 100.
    Violating these assumptions could lead to reaction quotients constantly overshooting their equilibrium.
    """

    def __init__(
        self,
        molecules: list[Molecule],
        reactions: list[tuple[list[Molecule], list[Molecule]]],
        abs_temp: float = 310.0,
        km_range: tuple[float, float] = (1e-2, 100.0),
        vmax_range: tuple[float, float] = (1e-3, 100.0),
        device: str = "cpu",
        scalar_enc_size: int = 64 - 3,
        vector_enc_size: int = 4096 - 3 * 64,
    ):
        self.abs_temp = abs_temp
        self.device = device
        self.abs_temp = abs_temp

        self.mol_energies = self._tensor_from([d.energy for d in molecules] * 2)
        self.mol_2_mi = {d: i for i, d in enumerate(molecules)}
        self.mi_2_mol = {v: k for k, v in self.mol_2_mi.items()}
        self.molecules = molecules

        # working cell params
        n_signals = 2 * len(molecules)
        self.Ke = self._tensor(0, 0)
        self.Kmf = self._tensor(0, 0)
        self.Kmb = self._tensor(0, 0)
        self.Kmr = self._tensor(0, 0)
        self.Vmax = self._tensor(0, 0)
        self.N = self._tensor(0, 0, n_signals)
        self.Nf = self._tensor(0, 0, n_signals)
        self.Nb = self._tensor(0, 0, n_signals)
        self.Ai = self._tensor(0, 0, n_signals)
        self.Aa = self._tensor(0, 0, n_signals)

        # the domain specifications return 4 indexes
        # idx 0-2 are 1-codon idxs for scalars (n=64)
        # idx3 is a 2-codon idx for vetors (n=4096)

        self.km_map = _LogNormWeightMapFact(
            max_token=scalar_enc_size,
            weight_range=km_range,
            device=device,
        )

        self.vmax_map = _LogNormWeightMapFact(
            max_token=scalar_enc_size,
            weight_range=vmax_range,
            device=device,
        )

        self.sign_map = _SignMapFact(max_token=scalar_enc_size, device=device)

        self.reaction_map = _ReactionMapFact(
            molmap=self.mol_2_mi,
            reactions=reactions,
            max_token=vector_enc_size,
            device=device,
        )

        self.transport_map = _TransporterMapFact(
            n_molecules=len(molecules), max_token=vector_enc_size, device=device
        )

        self.effector_map = _RegulatoryMapFact(
            n_molecules=len(molecules), max_token=vector_enc_size, device=device
        )

        # derive inverse maps for genome generation
        self.km_2_idxs = self.km_map.inverse()
        self.vmax_2_idxs = self.vmax_map.inverse()
        self.sign_2_idxs = self.sign_map.inverse()
        self.trnsp_2_idxs = self.transport_map.inverse(molecules=molecules)
        self.regul_2_idxs = self.effector_map.inverse(molecules=molecules)
        self.catal_2_idxs = self.reaction_map.inverse(
            molmap=self.mol_2_mi, reactions=reactions, n_signals=n_signals
        )

    def get_proteome(
        self,
        proteome: list[ProteinSpecType],
    ) -> list[Protein]:
        """
        Calculate cell parameters for a single proteome and return it as
        a list of proteins

        Arguments:
            proteome: proteome which should be calculated

        Retruns:
            A list of objects that represent proteins with domains and their
            specifications.
        """
        # get proteome tensors
        dom_types, idxs0, idxs1, idxs2, idxs3 = self._collect_proteome_idxs(
            proteomes=[[d[0] for d in proteome]]
        )

        # identify domain types
        # 1=catalytic, 2=transporter, 3=regulatory
        is_catal = dom_types == 1  # (c,p,d)
        is_trnsp = dom_types == 2  # (c,p,d)
        is_reg = dom_types == 3  # (c,p,d)

        # map indices of domain specifications to concrete values
        # idx0 is a 2-codon index specific for every domain type (n=4096)
        # idx1-3 are 1-codon used for the floats (n=64)
        # some values are not defined for certain domain types
        # setting their indices to 0 lets them map to empty values (0-vector, NaN)
        catal_lng = (is_catal).long()
        trnsp_lng = (is_trnsp).long()
        reg_lng = (is_reg).long()
        not_reg_lng = (~is_reg).long()

        # idxs 0-2 are 1-codon indexes used for scalars (n=64 (- stop codons))
        Vmax_d = self.vmax_map(idxs0 * not_reg_lng)  # float (c,p,d)
        Km_d = self.km_map(idxs1)  # float (c,p,d)
        signs = self.sign_map(idxs2)  # float (c,p,d)

        # idx3 is a 2-codon index used for vectors (n=4096 (- stop codons))
        reacts = self.reaction_map(idxs3 * catal_lng)  # float (c,p,d,s)
        trnspts = self.transport_map(idxs3 * trnsp_lng)  # float (c,p,d,s)
        effectors = self.effector_map(idxs3 * reg_lng)  # float (c,p,d,s)

        # N (c, p, d, s)
        N_d = torch.einsum("cpds,cpd->cpds", (reacts + trnspts), signs)

        # A (c, p, d, s)
        A_d = torch.einsum("cpds,cpd->cpds", effectors, signs)

        Nf_d = torch.where(N_d < 0.0, -N_d, 0.0)
        Nb_d = torch.where(N_d > 0.0, N_d, 0.0)
        mols = self.molecules
        n_mols = len(mols)

        # looping through list of lists while accessing tensors
        # should be ok because in _get_proteome_tensors none are filtered
        prots: list[Protein] = []
        for pi, protein_spec in enumerate(proteome):
            doms: list[Domain] = []
            for di, dom_spec in enumerate(protein_spec[0]):
                kwargs = {
                    "start": dom_spec[1],
                    "end": dom_spec[2],
                }

                # catalytic domain (N has positive and negative integers)
                if dom_types[0][pi][di].item() == 1:
                    lft_ns = Nf_d[0][pi][di].int().tolist()
                    rgt_ns = Nb_d[0][pi][di].int().tolist()
                    lfts = [m for m, n in zip(mols, lft_ns) for _ in range(n)]
                    rgts = [m for m, n in zip(mols, rgt_ns) for _ in range(n)]
                    if len(lfts) > 0:
                        doms.append(
                            CatalyticDomain(
                                reaction=(lfts, rgts),
                                km=Km_d[0][pi][di].item(),
                                vmax=Vmax_d[0][pi][di].item(),
                                **kwargs,
                            )
                        )

                # transporter domain (N has one +1 and one -1)
                if dom_types[0][pi][di].item() == 2:
                    mis = torch.argwhere(N_d[0][pi][di] != 0).int().flatten().tolist()
                    doms.append(
                        TransporterDomain(
                            molecule=self.mi_2_mol[min(mis)],
                            km=Km_d[0][pi][di].item(),
                            vmax=Vmax_d[0][pi][di].item(),
                            is_exporter=bool(N_d[0][pi][di][min(mis)] < 0),
                            **kwargs,
                        )
                    )

                # regulatory domain (A has one +1 or -1)
                if dom_types[0][pi][di].item() == 3:
                    mi = int(torch.argwhere(A_d[0][pi][di] != 0)[0].item())
                    if mi in self.mi_2_mol:
                        is_trnsm = False
                        mol = self.mi_2_mol[mi]
                    else:
                        is_trnsm = True
                        mol = self.mi_2_mol[mi - n_mols]
                    doms.append(
                        RegulatoryDomain(
                            effector=mol,
                            km=Km_d[0][pi][di].item(),
                            is_inhibiting=bool((A_d[0][pi][di][mi] == -1).item()),
                            is_transmembrane=is_trnsm,
                            **kwargs,
                        )
                    )

            if len(doms) > 0:
                prots.append(
                    Protein(
                        domains=doms,
                        cds_start=proteome[pi][1],
                        cds_end=proteome[pi][2],
                        is_fwd=proteome[pi][3],
                    )
                )

        return prots

    def set_cell_params(
        self,
        cell_idxs: list[int],
        proteomes: list[list[list[DomainSpecType]]],
    ):
        """
        Calculate and set cell parameters for new proteomes

        Arguments:
            cell_idxs: Indexes of cells which proteomes belong to
            proteomes: list of proteomes which should be calculated and set

        Proteomes must be represented as a list (proteomes) of lists (proteins)
        of lists (domains) which each carry tuples. These tuples are domain specifications
        that are derived by [Genetics][magicsoup.genetics.Genetics].
        These are indices which will be mapped to concrete values
        (molecule species, Km, Vmax, reactions, ...).
        """
        # get proteome tensors
        dom_types, idxs0, idxs1, idxs2, idxs3 = self._collect_proteome_idxs(
            proteomes=proteomes
        )

        # identify domain types
        # 1=catalytic, 2=transporter, 3=regulatory
        is_catal = dom_types == 1  # (c,p,d)
        is_trnsp = dom_types == 2  # (c,p,d)
        is_reg = dom_types == 3  # (c,p,d)

        # map indices of domain specifications to concrete values
        # idx0 is a 2-codon index specific for every domain type (n=4096)
        # idx1-3 are 1-codon used for the floats (n=64)
        # some values are not defined for certain domain types
        # setting their indices to 0 lets them map to empty values (0-vector, NaN)
        catal_lng = (is_catal).long()
        trnsp_lng = (is_trnsp).long()
        reg_lng = (is_reg).long()
        not_reg_lng = (~is_reg).long()

        # idxs 0-2 are 1-codon indexes used for scalars (n=64 (- stop codons))
        Vmaxs = self.vmax_map(idxs0 * not_reg_lng)  # float (c,p,d)
        Kms = self.km_map(idxs1)  # float (c,p,d)
        signs = self.sign_map(idxs2)  # float (c,p,d)

        # idx3 is a 2-codon index used for vectors (n=4096 (- stop codons))
        reacts = self.reaction_map(idxs3 * catal_lng)  # float (c,p,d,s)
        trnspts = self.transport_map(idxs3 * trnsp_lng)  # float (c,p,d,s)
        effectors = self.effector_map(idxs3 * reg_lng)  # float (c,p,d,s)

        # Vmax are averaged over domains
        # undefined Vmax enries are NaN and are ignored by nanmean
        self.Vmax[cell_idxs] = Vmaxs.nanmean(dim=2).nan_to_num(0.0)

        # effector vectors are summed up over domains
        # undefined vectors are all 0s
        A_d = torch.einsum("cpds,cpd->cpds", effectors, signs)

        # A for inhibitors and activators are distinguished
        self.Ai[cell_idxs] = torch.where(A_d < 0.0, -A_d, 0.0).sum(dim=2)
        self.Aa[cell_idxs] = torch.where(A_d > 0.0, A_d, 0.0).sum(dim=2)

        # Kms of regulatory domains are aggregated for effectors
        # Kms from other domains are ignored using NaNs and nanmean
        self.Kmr[cell_idxs] = (
            torch.where(is_reg, Kms, torch.nan).nanmean(dim=2).nan_to_num(0.0)
        )

        # reaction stoichiometry N is derived from transporter and catalytic vectors
        # vectors for regulatory domains or emptpy proteins are all 0s
        N_d = torch.einsum("cpds,cpd->cpds", (reacts + trnspts), signs)
        N = N_d.sum(dim=2)
        self.N[cell_idxs] = N

        # N for forward and backward reactions is distinguished
        # to not loose molecules like co-facors whose net N would become 0
        self.Nf[cell_idxs] = torch.where(N_d < 0.0, -N_d, 0.0).sum(dim=2)
        self.Nb[cell_idxs] = torch.where(N_d > 0.0, N_d, 0.0).sum(dim=2)

        # Kms of catalytic and transporter domains are aggregated
        # Kms from other domains are ignored using NaNs and nanmean
        Kmn = torch.where(~is_reg, Kms, torch.nan).nanmean(dim=2).nan_to_num(0.0)

        # energies define Ke which defines Ke = Kmf/Kmb
        # extreme energies can create Inf or 0.0, avoid them with clamp
        E = torch.einsum("cps,s->cp", N, self.mol_energies)
        Ke = torch.exp(-E / self.abs_temp / GAS_CONSTANT).clamp(_EPS, _MAX)
        self.Ke[cell_idxs] = Ke

        # Km is sampled between a defined range
        # exessively small Km can create numerical instability
        # thus, sampled Km should define the smaller Km of Ke = Kmf/Kmb
        # Ke>=1  => Kmf=Km,         Kmb=Ke*Km
        # Ke<1   => Kmf=Km/Ke,      Kmb=Km
        # this operation can create again Inf or 0.0, avoided with clamp
        # this effectively limits Ke around 1e38
        is_fwd = Ke >= 1.0
        self.Kmf[cell_idxs] = torch.where(is_fwd, Kmn, Kmn / Ke).clamp(_EPS, _MAX)
        self.Kmb[cell_idxs] = torch.where(is_fwd, Kmn * Ke, Kmn).clamp(_EPS, _MAX)

    def unset_cell_params(self, cell_idxs: list[int]):
        """
        Unset cell parameters (Vmax, Km, ...) for cells with empty
        or non-viable proteomes.

        Arguments:
            cell_idxs: Indexes of cells
        """
        self.N[cell_idxs] = 0.0
        self.Nf[cell_idxs] = 0.0
        self.Nb[cell_idxs] = 0.0
        self.Ai[cell_idxs] = 0.0
        self.Aa[cell_idxs] = 0.0
        self.Ke[cell_idxs] = 0.0
        self.Kmf[cell_idxs] = 0.0
        self.Kmb[cell_idxs] = 0.0
        self.Kmr[cell_idxs] = 0.0
        self.Vmax[cell_idxs] = 0.0

    def copy_cell_params(self, from_idxs: list[int], to_idxs: list[int]):
        """
        Copy paremeters from a list of cells to another list of cells

        Arguments:
            from_idxs: List of cell indexes to copy from
            to_idxs: List of cell indexes to copy to

        `from_idxs` and `to_idxs` must have the same length.
        They refer to the same cell indexes as in `world.cell_genomes`.
        """
        self.Ke[to_idxs] = self.Ke[from_idxs]
        self.Kmf[to_idxs] = self.Kmf[from_idxs]
        self.Kmb[to_idxs] = self.Kmb[from_idxs]
        self.Kmr[to_idxs] = self.Kmr[from_idxs]
        self.Vmax[to_idxs] = self.Vmax[from_idxs]
        self.N[to_idxs] = self.N[from_idxs]
        self.Nf[to_idxs] = self.Nf[from_idxs]
        self.Nb[to_idxs] = self.Nb[from_idxs]
        self.Ai[to_idxs] = self.Ai[from_idxs]
        self.Aa[to_idxs] = self.Aa[from_idxs]

    def remove_cell_params(self, keep: torch.Tensor):
        """
        Remove cells from cell params

        Arguments:
            keep: Bool tensor (c,) which is true for every cell that should not be removed
                and false for every cell that should be removed.

        `keep` must have the same length as `world.cell_genomes`.
        The indexes on `keep` reflect the indexes in `world.cell_genomes`.
        """
        self.Ke = self.Ke[keep]
        self.Kmf = self.Kmf[keep]
        self.Kmb = self.Kmb[keep]
        self.Kmr = self.Kmr[keep]
        self.Vmax = self.Vmax[keep]
        self.N = self.N[keep]
        self.Nf = self.Nf[keep]
        self.Nb = self.Nb[keep]
        self.Ai = self.Ai[keep]
        self.Aa = self.Aa[keep]

    def increase_max_cells(self, by_n: int):
        """
        Increase the cell dimension of all cell parameters

        Arguments:
            by_n: By how many rows to increase the cell dimension
        """
        self.Ke = self._expand_c(t=self.Ke, n=by_n)
        self.Kmf = self._expand_c(t=self.Kmf, n=by_n)
        self.Kmb = self._expand_c(t=self.Kmb, n=by_n)
        self.Kmr = self._expand_c(t=self.Kmr, n=by_n)
        self.Vmax = self._expand_c(t=self.Vmax, n=by_n)
        self.N = self._expand_c(t=self.N, n=by_n)
        self.Nf = self._expand_c(t=self.Nf, n=by_n)
        self.Nb = self._expand_c(t=self.Nb, n=by_n)
        self.Ai = self._expand_c(t=self.Ai, n=by_n)
        self.Aa = self._expand_c(t=self.Aa, n=by_n)

    def increase_max_proteins(self, max_n: int):
        """
        Increase the protein dimension of all cell parameters

        Arguments:
            max_n: The maximum number of rows required in the protein dimension
        """
        n_prots = int(self.N.size(1))
        if max_n > n_prots:
            by_n = max_n - n_prots
            self.Ke = self._expand_p(t=self.Ke, n=by_n)
            self.Kmf = self._expand_p(t=self.Kmf, n=by_n)
            self.Kmb = self._expand_p(t=self.Kmb, n=by_n)
            self.Kmr = self._expand_p(t=self.Kmr, n=by_n)
            self.Vmax = self._expand_p(t=self.Vmax, n=by_n)
            self.N = self._expand_p(t=self.N, n=by_n)
            self.Nf = self._expand_p(t=self.Nf, n=by_n)
            self.Nb = self._expand_p(t=self.Nb, n=by_n)
            self.Ai = self._expand_p(t=self.Ai, n=by_n)
            self.Aa = self._expand_p(t=self.Aa, n=by_n)

    def integrate_signals(self, X: torch.Tensor) -> torch.Tensor:
        """
        Simulate protein work by integrating all signals.

        Arguments:
            X: Tensor of every signal in every cell (c, s). Must all be >= 0.0.

        Returns:
            New tensor of the same shape which represents the updated signals for every cell.

        The order of cells in `X` is the same as in `world.cell_genomes`
        The order of signals is first all intracellular molecule species in the same order as `chemistry.molecules`,
        then again all molecule species in the same order but this time describing extracellular molecule species.
        The number of intracellular molecules comes from `world.cell_molecules` for any particular cell.
        The number of extracellular molecules comes from `world.molecule_map` from the pixel the particular cell currently lives on.
        """
        # calculate in multiple parts while reducing velocity
        # trim factors get increasingly smaller to make it more likely to approach the
        # reaction equilibrium if it was overshot in the previous part
        trim_factors = (0.7, 0.2, 0.1)
        for trim in trim_factors:
            X = self._integrate_signals_part(
                adj_vmax=(self.Vmax * trim).clamp(0.0), X0=X
            )
        return X

    def _integrate_signals_part(
        self, adj_vmax: torch.Tensor, X0: torch.Tensor
    ) -> torch.Tensor:
        V = self._get_velocities(X=X0, Vmax=adj_vmax)  # (c,p)

        # NV to keep synthesis and deconstruction separate
        NV = torch.einsum("cps,cp->cps", self.N, V)  # (c,p,s)

        # adjust NV downward to avoid negative concentrations
        NV_adj = self._get_negative_adjusted_nv(NV=NV, X=X0)  # (c,p,s)
        X1 = X0 + NV_adj.sum(1)  # (c,s)
        X1[X1 < 0.0] = 0.0  # small floating point errors can lead to -1e-7

        # adjust X downward to avoid Q overshooting Ke
        X1_adj = self._get_equilibrium_adjusted_x(X0=X0, X1=X1, NV=NV_adj, V=V)

        return X1_adj

    def _get_velocities(self, X: torch.Tensor, Vmax: torch.Tensor) -> torch.Tensor:
        # catalytic activity

        # signals are aggregated for forward and backward reaction
        # proteins that had no involved catalytic region should not be active
        kf, f_prots = self._multiply_signals(X=X, N=self.Nf)
        kf /= self.Kmf
        kf[~f_prots] = 0.0  # non-involved proteins are not active
        kf[kf.isinf()] = _MAX

        kb, b_prots = self._multiply_signals(X=X, N=self.Nb)
        kb /= self.Kmb
        kb[~b_prots] = 0.0  # non-involved proteins are not active
        kb[kb.isinf()] = _MAX

        # custom reversible MM equation
        a_cat = (kf - kb) / (1 + kf + kb)  # (c,p)

        # inhibitor activity
        xxi, i_prots = self._multiply_signals(X=X, N=self.Ai)
        a_inh = xxi / (xxi + self.Kmr)
        a_inh[~i_prots] = 0.0  # proteins without inhibitor should be active

        # activator activity
        xxa, a_prots = self._multiply_signals(X=X, N=self.Aa)
        a_act = xxa / (xxa + self.Kmr)
        a_act[~a_prots] = 1.0  # proteins without activator should be active

        # velocity from activities
        # as well as trimming factor of n_computations
        V = a_cat * Vmax * (1 - a_inh) * a_act  # (c,p)

        # Kms can be close to Inf, they can create Infs
        # thus result velocity should be clamped again
        return V.clamp(_MIN, _MAX)

    def _get_equilibrium_adjusted_x(
        self,
        X0: torch.Tensor,
        X1: torch.Tensor,
        NV: torch.Tensor,
        V: torch.Tensor,
    ) -> torch.Tensor:
        # adjust NV downwards so Q does not overshoot Ke
        # must calculate X1 and Q1 to do adjustments, X1 is returned
        # all proteins influence each others Q, so iterative heuristic
        # increments to an adjustment factor in steps
        increments = (0.5, 0.25, 0.125, 0.0625)

        # stop adjusting if Ke ~= Q
        upper_thresh = 1.5
        lower_thresh = 1 / 1.5

        has_impact = V.abs() > 0.1
        is_fwd = V > 0.0

        # running factors for adjusting NV
        F = torch.ones_like(V)  # (c,p)
        for increment in increments:
            # calculate expected quotient Q1 and compare to Ke (t1)
            Q1 = self._get_quotient(X=X1)  # (c,p)
            QKe = Q1 / self.Ke

            # fwd reaction: Q -> Ke from below, QKe > 1 is overshoot
            # bwd reaction: Q -> Ke from above, QKe < 1 is overshoot
            # reactions cant be adjusted higher than 1.0

            v_too_low = torch.where(is_fwd, QKe < lower_thresh, QKe > upper_thresh)
            v_too_low[is_fwd & (F == 1.0)] = False

            v_too_high = torch.where(is_fwd, QKe > upper_thresh, QKe < lower_thresh)
            v_too_high[~is_fwd & (F == 0.0)] = False

            # mask for which velocities can be adjusted
            if not torch.any((v_too_low | v_too_high) & has_impact):
                return X1

            # adjust F upwards/downwards
            F[v_too_high] -= increment
            F[v_too_low] += increment
            F[F > 1.0] = 1.0
            F[F < 0.0] = 0.0

            # calculate new X1
            X1 = X0 + torch.einsum("cps,cp->cs", NV, F)  # (c,s)
            X1[X1 < 0.0] = 0.0

        return X1

    def _get_negative_adjusted_nv(
        self, NV: torch.Tensor, X: torch.Tensor
    ) -> torch.Tensor:
        # how does each protein have to be slowed down for each signal
        # so not more signal than available is removed
        F = X / (-NV).clamp(min=0.0).sum(1)  # (c,s)
        F[F > 1.0] = 1.0

        # which signals are removed by each protein
        M_rm = NV < 0.0  # (c,p,s)

        # factor for slowing down each protein for each signal
        F_prots = torch.einsum("cps,cs->cps", M_rm.float(), F)
        F_prots[~M_rm] = 1.0  # (c,p,s)

        # how much can each protein be active at most
        F_min = F_prots.min(dim=2).values  # (c,p)

        return torch.einsum("cps,cp->cps", NV, F_min)

    def _get_quotient(self, X: torch.Tensor) -> torch.Tensor:
        xx_prod, prod_prots = self._multiply_signals(X=X, N=self.Nb)
        xx_prod[~prod_prots] = 0.0
        xx_prod[xx_prod.isinf()] = _MAX

        xx_subs, subs_prots = self._multiply_signals(X=X, N=self.Nf)
        xx_subs[~subs_prots] = 0.0
        xx_subs[xx_subs.isinf()] = _MAX

        # note: x/0=Inf, 0/x=0, 0/0=nan
        # 0.0 and Inf is avoided as it is in Ke calculation
        return (xx_prod / xx_subs).clamp(min=_EPS, max=_MAX).nan_to_num(1.0)

    def _multiply_signals(
        self, X: torch.Tensor, N: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        # calculate torch.prod(torch.pow(x, n), 2)
        # consider:
        # - some proteins are not involved at all (their Ns are all 0)
        # - some are involved but the signal 0 (required X is 0)
        # - 0^n=0 but x^0=1
        M = N > 0

        # expand to p while keeping uninvolved 0
        x = torch.einsum("cps,cs->cps", M, X)

        # stray 1s created here
        xx = torch.prod(torch.pow(x, N), 2)

        # Infs could have been created, MAX is still >e2 away from Inf
        # also I'm not sure how but somehow small negative values
        # (<1e-7) can be produced (I suspect floating point errors)
        xx[xx.isnan()] = 0.0
        xx[xx < 0.0] = 0.0
        xx[xx.isinf()] = _MAX

        # mask (c,p) which proteins are active must be returned
        return xx, M.sum(2) > 0.0

    def _collect_proteome_idxs(
        self,
        proteomes: list[list[list[DomainSpecType]]],
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        n_prots = self.N.size(1)
        n_doms = max(len(dd) for d in proteomes for dd in d)
        empty_seq = [0] * n_doms

        c_dts = []
        c_idxs0 = []
        c_idxs1 = []
        c_idxs2 = []
        c_idxs3 = []
        for proteins in proteomes:
            p_dts = []
            p_idxs0 = []
            p_idxs1 = []
            p_idxs2 = []
            p_idxs3 = []
            for doms in proteins:
                d_dts = []
                d_idxs0 = []
                d_idxs1 = []
                d_idxs2 = []
                d_idxs3 = []
                for (dt, i0, i1, i2, i3), *_ in doms:
                    d_dts.append(dt)
                    d_idxs0.append(i0)
                    d_idxs1.append(i1)
                    d_idxs2.append(i2)
                    d_idxs3.append(i3)
                d_pad = n_doms - len(d_idxs0)
                p_dts.append(d_dts + [0] * d_pad)
                p_idxs0.append(d_idxs0 + [0] * d_pad)
                p_idxs1.append(d_idxs1 + [0] * d_pad)
                p_idxs2.append(d_idxs2 + [0] * d_pad)
                p_idxs3.append(d_idxs3 + [0] * d_pad)

            p_pad = n_prots - len(p_idxs0)
            c_dts.append(p_dts + [empty_seq] * p_pad)
            c_idxs0.append(p_idxs0 + [empty_seq] * p_pad)
            c_idxs1.append(p_idxs1 + [empty_seq] * p_pad)
            c_idxs2.append(p_idxs2 + [empty_seq] * p_pad)
            c_idxs3.append(p_idxs3 + [empty_seq] * p_pad)

        dom_types = self._tensor_from(c_dts)  # long (c,p,d)
        idxs0 = self._tensor_from(c_idxs0)  # long (c,p,d)
        idxs1 = self._tensor_from(c_idxs1)  # long (c,p,d)
        idxs2 = self._tensor_from(c_idxs2)  # long (c,p,d)
        idxs3 = self._tensor_from(c_idxs3)  # long (c,p,d)
        return dom_types, idxs0, idxs1, idxs2, idxs3

    def _expand_c(self, t: torch.Tensor, n: int) -> torch.Tensor:
        size = t.size()
        zeros = self._tensor(n, *size[1:])
        return torch.cat([t, zeros], dim=0)

    def _expand_p(self, t: torch.Tensor, n: int) -> torch.Tensor:
        size = t.size()
        zeros = self._tensor(size[0], n, *size[2:])
        return torch.cat([t, zeros], dim=1)

    def _tensor(self, *args) -> torch.Tensor:
        return torch.zeros(*args).to(self.device)

    def _tensor_from(self, d: Any) -> torch.Tensor:
        return torch.tensor(d).to(self.device)



import torch
import random
import magicsoup as ms
from magicsoup.examples.wood_ljungdahl import MOLECULES, REACTIONS


def _get_kinetics() -> ms.Kinetics:
    chemistry = ms.Chemistry(molecules=MOLECULES, reactions=REACTIONS)
    kinetics = ms.Kinetics(
        molecules=chemistry.molecules,
        reactions=chemistry.reactions,
        abs_temp=310,
        scalar_enc_size=61,
        vector_enc_size=3904,
    )
    return kinetics


def test_cell_params_are_always_set_reproduceably():
    n_cells = 100
    n_tries = 10

    for i in range(n_tries):
        proteomes = []
        for _ in range(n_cells):
            proteins = []
            for _ in range(random.choice(list(range(20)))):
                domains = []
                for _ in range(random.choice([1, 1, 2])):
                    domtype = random.choice([1, 2, 3])
                    i0 = random.choice(list(range(61)))
                    i1 = random.choice(list(range(61)))
                    i2 = random.choice(list(range(61)))
                    i3 = random.choice(list(range(3904)))
                    domains.append(((domtype, i0, i1, i2, i3), 1, 2))
                proteins.append(domains)
            proteomes.append(proteins)

        n_max_prots = max(len(d) for d in proteomes)
        idxs = list(range(n_cells))

        kinetics = _get_kinetics()
        kinetics.increase_max_cells(by_n=n_cells)
        kinetics.increase_max_proteins(max_n=n_max_prots)
        kinetics.set_cell_params(cell_idxs=idxs, proteomes=proteomes)

        N_orig = kinetics.N.clone()
        Nb_orig = kinetics.Nb.clone()
        Nf_orig = kinetics.Nf.clone()
        Ke_orig = kinetics.Ke.clone()
        Kmf_orig = kinetics.Kmf.clone()
        Kmb_orig = kinetics.Kmb.clone()
        Kmr_orig = kinetics.Kmr.clone()
        Vmax_orig = kinetics.Vmax.clone()

        kinetics.remove_cell_params(keep=torch.full((n_cells,), False))
        kinetics.increase_max_cells(by_n=n_cells)
        kinetics.increase_max_proteins(max_n=n_max_prots)
        kinetics.set_cell_params(cell_idxs=idxs, proteomes=proteomes)

        assert torch.equal(kinetics.N, N_orig), i
        assert torch.equal(kinetics.Nb, Nb_orig), i
        assert torch.equal(kinetics.Nf, Nf_orig), i
        assert torch.equal(kinetics.Ke, Ke_orig), i
        assert torch.equal(kinetics.Kmf, Kmf_orig), i
        assert torch.equal(kinetics.Kmb, Kmb_orig), i
        assert torch.equal(kinetics.Kmr, Kmr_orig), i
        assert torch.equal(kinetics.Vmax, Vmax_orig), i


def test_random_kinetics_stay_zero():
    # Typical reasons for cells creating signals from 0:
    # 1. when using exp(ln(x)) 1s can accidentally be created
    # 2. when doing a^b 1s can be created (0^1=0 but 0^0=1)
    n_cells = 100
    n_prots = 100
    n_steps = 1000

    kinetics = _get_kinetics()
    n_mols = len(MOLECULES) * 2

    # concentrations (c, s)
    X = torch.zeros(n_cells, n_mols).abs()

    # reactions (c, p, s)
    kinetics.N = torch.randint(low=-8, high=9, size=(n_cells, n_prots, n_mols)).float()
    kinetics.Nf = torch.where(kinetics.N < 0.0, -kinetics.N, 0.0)
    kinetics.Nb = torch.where(kinetics.N > 0.0, kinetics.N, 0.0)

    # max velocities (c, p)
    kinetics.Vmax = torch.randn(n_cells, n_prots).abs() * 100

    # allosterics (c, p, s)
    kinetics.Ai = torch.randint(low=0, high=5, size=(n_cells, n_prots, n_mols))
    kinetics.Aa = torch.randint(low=0, high=5, size=(n_cells, n_prots, n_mols))

    # affinities (c, p)
    Ke = torch.randn(n_cells, n_prots) * 100
    kinetics.Kmf = torch.randn(n_cells, n_prots).abs()
    kinetics.Kmb = kinetics.Kmf * Ke
    kinetics.Kmr = torch.randn(n_cells, n_prots).abs()
    kinetics.Ke = kinetics.Kmb / kinetics.Kmf

    # test
    for _ in range(n_steps):
        X = kinetics.integrate_signals(X=X)
        assert X.min() == 0.0
        assert X.max() == 0.0


def test_random_kinetics_dont_explode():
    # interaction terms can overflow float32 when they become too big
    # i.g. exponents too high and many substrates
    # this will create infinites which will eventually become NaN
    n_cells = 100
    n_prots = 100
    n_steps = 1000

    kinetics = _get_kinetics()
    n_mols = len(MOLECULES) * 2

    # concentrations (c, s)
    X = torch.randn(n_cells, n_mols).abs().clamp(max=1.0) * 100

    # reactions (c, p, s)
    kinetics.N = torch.randint(low=-8, high=9, size=(n_cells, n_prots, n_mols)).float()
    kinetics.Nf = torch.where(kinetics.N < 0.0, -kinetics.N, 0.0)
    kinetics.Nb = torch.where(kinetics.N > 0.0, kinetics.N, 0.0)

    # max velocities (c, p)
    kinetics.Vmax = torch.randn(n_cells, n_prots).abs().clamp(max=1.0) * 100

    # allosterics (c, p, s)
    kinetics.Ai = torch.randint(low=0, high=5, size=(n_cells, n_prots, n_mols))
    kinetics.Aa = torch.randint(low=0, high=5, size=(n_cells, n_prots, n_mols))

    # affinities (c, p)
    Ke = torch.randn(n_cells, n_prots) * 100
    kinetics.Kmf = torch.randn(n_cells, n_prots).abs().clamp(0.001)
    kinetics.Kmb = (kinetics.Kmf * Ke).clamp(0.001)
    kinetics.Kmr = torch.randn(n_cells, n_prots).abs().clamp(0.001)
    kinetics.Ke = kinetics.Kmb / kinetics.Kmf

    # test
    for _ in range(n_steps):
        X = kinetics.integrate_signals(X=X)
        assert not torch.any(X < 0.0), X[X < 0.0].min()
        assert not torch.any(X.isnan()), X.isnan().sum()
        assert torch.all(X.isfinite()), ~X.isfinite().sum()
        assert torch.all(X < 10_000), X[X >= 10_000]
